# Copyright (C) 2000 Bruce Guenter <bruceg@em.ca>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import re

import context
import parser

path = [ '' ]

###############################################################################
# Node types

class Node:
	pass

def indent(level): return ' - ' * level

class Body(Node):
	def __init__(self):
		self.list = [ ]
	def append(self, node):
		self.list.append(node)
	def showtree(self, level):
		print indent(level), "Body"
		for item in self.list:
			item.showtree(level+1)
	def execute(self, context):
		for item in self.list:
			item.execute(context)

class Section(Node):
	'''Encapsulates an optionally formatted text section'''
	def __init__(self, string, formatted=0):
		self.section = string
	def showtree(self, level):
		print indent(level), "Section(%d bytes)" % len(self.section)
	def execute(self, context):
		return context.out.write(self.section % context)

class End(Node):
	'''Ends a block node, like Start, Foreach, and If'''
	rx = re.compile(r'^end(\s+.*|)$')
	def __init__(self, ignored):
		pass

class Exit(Node,Exception):
	'''Causes parsing to stop'''
	rx = re.compile(r'^exit$')
	def __init__(self):
		pass
	def showtree(self, level):
		print indent(level), "Exit"
	def execute(self, context):
		raise self

class If(Node):
	rx = re.compile(r'^if\s+(.+)$')
	def __init__(self, expr):
		self.expr = expr
		self.body = None
		self.else_node = None
		self.elseifs = [ ]
	def append_elseif(self, node):
		self.elseifs.append(node)
	def showtree(self, level):
		print indent(level), "If", self.expr
		self.body.showtree(level+1)
		for elseif in self.elseifs:
			elseif.showtree(level)
		if self.else_node:
			self.else_node.showtree(level)
	def execute(self, context):
		if context.eval(self.expr):
			return self.body.execute(context)
		for elseif in self.elseifs:
			if elseif.execute(context):
				return
		if self.else_node:
			return self.else_node.execute(context)

class Else(Node):
	'''Negates an If clause'''
	rx = re.compile(r'^else$')
	def __init__(self):
		self.body = None
	def showtree(self, level):
		print indent(level), "Else"
		self.body.showtree(level+1)
	def execute(self, context):
		return self.body.execute(context)

class ElseIf(Node):
	'''Additional clauses to an If clause'''
	rx = re.compile(r'^else\s+if\s+(.+)$')
	def __init__(self, expr):
		self.expr = expr
		self.body = None
	def showtree(self, level):
		print indent(level), "ElseIf", self.expr
		self.body.showtree(level+1)
	def execute(self, context):
		if context.eval(self.expr):
			self.body.execute(context)
			return 1
		return None

class For(Node):
	'''Iterates over each item in a list'''
	rx = re.compile(r'^for\s+(\w+)\s+in\s+(.+)$')
	def __init__(self, var, expr):
		self.var = var
		self.expr = expr
		self.body = None
	def showtree(self, level):
		print indent(level), "For", self.var, "in", self.expr
		self.body.showtree(level + 1)
	def execute(self, context):
		context.push()
		list = context.eval(self.expr)
		for item in list:
			context[self.var] = item
			self.body.execute(context)
		context.pop()

class Foreach(Node):
	'''Iterates over each item in a list of dicts'''
	rx = re.compile(r'^foreach\s+(.+)$')
	def __init__(self, expr):
		self.expr = expr
		self.body = None
	def showtree(self, level):
		print indent(level), "Foreach", self.expr
		self.body.showtree(level + 1)
	def execute(self, context):
		context.push()
		list = context.eval(self.expr)
		for item in list:
			context.update(item)
			self.body.execute(context)
		context.pop()

_include_cache = { }
class Include(Node):
	'''Reads in and parses another file'''
	rx = re.compile(r'^include\s+(.+)$')
	def __init__(self, expr):
		self.expr = expr
	def showtree(self, level):
		print indent(level), "Include", self.expr
	def execute(self, context):
		filename = context.eval(self.expr)
		try:
			tree = _include_cache[filename]
		except KeyError:
			tree = None
			for prefix in path:
				fullname = os.path.join(prefix, filename)
				try:
					content = open(fullname, 'r').read()
					tree = parser.parse(content)
					break
				except IOError:
					pass
			if tree is None:
				raise IOError, "Could not find file '%s'" % filename
		tree.execute(context)

class With(Node):
	'''Temporarily adds some symbols to the running context'''
	rx = re.compile(r'^with\s+(.+)$')
	def __init__(self, expr):
		self.expr = expr
		self.body = None
	def showtree(self, level):
		print indent(level), "With", self.expr
		self.body.showtree(level + 1)
	def execute(self, context):
		context.push()
		dict = context.eval(self.expr)
		context.update(dict)
		self.body.execute(context)
		context.pop()

special_node_types = [ Else, ElseIf, End, Exit, For, Foreach, If,
					   Include, With ]

class Expr(Node):
	def __init__(self, expr):
		self.expr = expr
	def showtree(self, level):
		print indent(level), "Expr", self.expr
	def execute(self, context):
		return context.execute(self.expr)
